# -*- coding: utf-8 -*-
from os.path import abspath, dirname, join

ROOT = abspath(dirname(__file__))

VALID_ONIX2_REF = join(ROOT, "valid_onix2_ref.xml")

VALID_ONIX3_REF = join(ROOT, "valid_onix3_ref.xml")
VALID_ONIX3_SHORT = join(ROOT, "valid_onix3_short.xml")

VALID_ONIX31_REF = join(ROOT, "valid_onix31_ref.xml")
VALID_ONIX31_SHORT = join(ROOT, "valid_onix31_short.xml")

VALID_ONIX3_REF_NS = join(ROOT, "valid_onix3_ref_ns.xml")

INVALID_ONIX3_REF = join(ROOT, "invalid_onix3_ref.xml")
INVALID_ONIX_ROOT = join(ROOT, "invalid_onix_root.xml")

VALID_GOOGLE_SAMPLE = join(ROOT, "Google Sample ONIX 2.1.xml")

VALID_GOOGLE_ONIX_30_SAMPLE = join(ROOT, "valid_google_onix_3.xml")
INVALID_GOOGLE_ONIX_30_SAMPLE = join(ROOT, "invalid_google_onix_3.xml")

WIN_CONSOLE_ISSUE = join(ROOT, "win_console_issue.xml")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import string
from os.path import basename, splitext
from datetime import datetime
from decimal import Decimal, InvalidOperation
import yaml
from lxml import objectify
import isbnlib
from onixcheck.models import OnixFile, Message


class OnixFix(object):
    """
    Validates Onix xml against custom yaml based subspecifications and
    provides usefull error and warning messages.

    :param onixfile: File obj with onix data or path to onix file
    :type onixfile: file or str
    :param profile: Path to yaml profile
    :type profile: str
    """

    def __init__(self, onixfile, profile):
        if hasattr(onixfile, "name"):
            self.basename = basename(onixfile.name)
        else:
            self.basename = basename(onixfile)

        if hasattr(onixfile, "seek"):
            onixfile.seek(0)

        self._onix_file = OnixFile(onixfile)
        self.onix = self._onix_file.xml_tree().getroot()

        # Get rid of namespaces for easier xpath search
        for elem in self.onix.getiterator():
            if not hasattr(elem.tag, "find"):
                continue
            i = elem.tag.find("}")
            if i >= 0:
                elem.tag = elem.tag[i + 1 :]
        objectify.deannotate(self.onix, cleanup_namespaces=True)

        self.spec = yaml.load(open(profile, "rb"), Loader=yaml.SafeLoader)

        name, ext = splitext(basename(profile))
        self.name = name.upper()
        self.urls_checked = set()
        self.errors = []
        self.warnings = []

    @property
    def messages(self):
        return self.errors + self.warnings

    def add_message(self, message, path, el=None, level="ERROR"):
        assert level in (
            "ERROR",
            "WARNING",
        )

        if el is not None and el.sourceline is not None:
            location = "%s" % el.sourceline
        else:
            location = path

        location = "%s:%s" % (self.basename, location)

        msg = Message(
            level=level,
            validator=self.name,
            location=location,
            message=message,
            error_type="SPECIFICATION ERROR",
        )
        if level == "ERROR":
            self.errors.append(msg)
        elif level == "WARNING":
            self.warnings.append(msg)

    def validate(self):
        """
        Validate Onix agianst Specfile

        :return list[Message]: list of error messages if any else empty list.
        """

        for path_or_keyword in self.spec:
            if path_or_keyword.startswith("/"):
                path = path_or_keyword.lstrip("/")
                specnode = self.spec[path_or_keyword]
                xmlnode = self.onix
                self.validate_block(path, specnode, xmlnode)
        return self.errors

    def validate_block(self, path, specnode, xmlnode):
        for pok in specnode:
            if pok in ("description", "example"):
                continue
            elif pok in ("required", "mulitple", "recommended"):
                func = getattr(self, "handle_" + pok)
                func(path, xmlnode, specnode.get(pok))
            elif pok.startswith("/"):
                sub_path = pok.lstrip("/")
                for node in xmlnode.xpath(path):
                    self.validate_block(sub_path, specnode[pok], node)
            else:
                func = getattr(self, "handle_" + pok)
                value = specnode.get(pok)
                for element in xmlnode.xpath(path):
                    func(path, element, value)

    def handle_required(self, path, el, val):
        if val:
            result = el.xpath(path)
            if len(result) == 0:
                msg = "Missing element %s within %s" % (path, el.tag)
                self.add_message(msg, path, el)

    def handle_requires_one_of(self, path, el, val):
        if val:
            results = []
            for elname in val:
                result = el.xpath(elname)
                if len(result) == 1:
                    results.append(result)
            if len(results) != 1:
                msg = "Exactly one of %s must be within %s" % (val, el.tag)
                self.add_message(msg, path)

    def handle_recommended(self, path, el, val):
        if val:
            result = el.xpath(path)
            if len(result) == 0:
                msg = "Missing recommended element %s within %s" % (path, el.tag)
                self.add_message(msg, path, level="WARNING")

    def handle_multiple(self, path, el, val):
        if not val:
            result = el.getparent().xpath(path)
            if len(result) > 1:
                msg = 'Only one "%s" allowed per "%s" found %s' % (path, el.getparent().tag, len(result))
                self.add_message(msg, path)

    def handle_content(self, path, el, val):
        if el.text != val:
            msg = 'Expected content "%s" does not match "%s"' % (val, el.text)
            self.add_message(msg, path, el)

    def handle_length(self, path, el, val):
        if len(el.text) != val:
            msg = 'Expected %s characters for "%s"' % (val, el.text)
            self.add_message(msg, path, el)

    def handle_choice(self, path, el, val):
        if el.text not in val:
            msg = '"%s" not one of "%s"' % (el.text, val)
            self.add_message(msg, path, el)

    def handle_dateformat(self, path, el, val):
        try:
            datetime.strptime(el.text, val).date()
        except ValueError as e:
            self.add_message(e.message, path, el)

    def handle_range(self, path, el, val):
        if not int(el.text) in range(*val):
            msg = '"%s" out of range' % el.text
            self.add_message(msg, path, el)

    def handle_type(self, path, el, val):
        assert val in ("int", "isbn13", "isbn", "decimal")
        getattr(self, "check_" + val)(path, el)

    def check_int(self, path, el):
        invalid = [c for c in el.text if c not in string.digits]
        if invalid:
            msg = "Not an Integer. Invalid chars %s." % invalid
            self.add_message(msg, path, el)

    def check_isbn13(self, path, el):
        if not isbnlib.is_isbn13(el.text):
            msg = "Invalid ISBN13: %s" % el.text
            self.add_message(msg, path, el)

    def check_isbn(self, path, el):
        if not isbnlib.is_isbn13(el.text) and not isbnlib.is_isbn10(el.text):
            msg = "Invalid ISBN: %s" % el.text
            self.add_message(msg, path, el)

    def check_decimal(self, path, el):
        try:
            Decimal(el.text)
        except InvalidOperation:
            msg = "Invalid decimal value %s" % el.text
            self.add_message(msg, path, el)


if __name__ == "__main__":
    from onixcheck import profiles, data

    of = OnixFix(data.VALID_GOOGLE_ONIX_30_SAMPLE, profiles.GOOGLE_ONIX_30)
    of.validate()

    for message in of.messages:
        print(message.short)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import re
from collections import namedtuple
from lxml import etree
from onixcheck import schema
from onixcheck.exeptions import OnixError, get_logger


log = get_logger()


class OnixFile(object):
    """Convenience file object wrapper.

    :param infile: File or path to file
    :type infile: file or str
    """

    def __init__(self, infile):
        self.infile = infile
        self.meta = OnixMeta.from_file(infile)
        if hasattr(infile, "seek"):
            self.infile.seek(0)

    def xml_tree(self):
        """
        Parse the infile with lxml and add the proper namespace if required.

        :return etree.ElementTree: An lxml ElementTree with proper namespace
        """
        if hasattr(self.infile, "seek"):
            self.infile.seek(0)

        tree = etree.parse(self.infile)

        if self.meta.namespaces:
            return tree

        log.debug("Adding namespaces to xml for validation")
        root = tree.getroot()
        ns_root = etree.Element(tree.docinfo.root_name, root.attrib, nsmap={None: self.meta.get_ns_string()})
        ns_root[:] = root[:]

        # Roundtrip to add namespace
        doc = etree.tostring(ns_root, encoding=tree.docinfo.encoding, xml_declaration=True, pretty_print=True)
        ns_tree = etree.fromstring(doc)
        return etree.ElementTree(ns_tree)

    def get_validator(self, schema_type="xsd"):
        """
        Create a matching validator for the ONIX file.

        :return etree._Validator:
        """
        parser = self.meta.SCHEMA_TYPE_PARSER_MAP[schema_type]
        return parser(file=self.meta.get_schema_file(schema_type=schema_type))


_BaseMeta = namedtuple("OnixMeta", "xml_version xml_encoding onix_version onix_style namespaces")


class OnixMeta(_BaseMeta):
    """Read and detect minimal ONIX file properties needed for validation.

    Onix XML files may or may not have `release` and `xmlns` attributes on
    their root element. OnixMeta.from_file(infile) will detect Onix Version
    and Style and also patch the root element with the appropriate namespace
    needed for validation.

    :param str xml_version: XML Version as str ("1.0").
    :param str xml_encoding: XML Encoding as str ("utf-8").
    :param str onix_version: Onix Version as string ("2.1", "3.0", "3.1")
    :param str onix_style: Onix Style as str ("short" or "reference")
    :param dict namespaces: dict of str with namspaces from the root element

    """

    #: ONIX Version 2.1
    V21 = "2.1"
    #: ONIX Version 3.0
    V30 = "3.0"
    #: ONIX Version 3.1
    V31 = "3.1"
    #: Short notation
    SHORT = "short"
    #: Reference notation
    REFERENCE = "reference"
    #: Schema Types
    XSD = "xsd"
    RNG = "rng"

    ONIX_VERSIONS = (V21, V30, V31)
    ONIX_STYLES = (SHORT, REFERENCE)

    SCHEMA_MAP = {
        (V21, SHORT, XSD): schema.O21_XSD_SHORT,
        (V21, REFERENCE, XSD): schema.O21_XSD_REFERENCE,
        (V30, SHORT, XSD): schema.O30_XSD_SHORT,
        (V30, SHORT, RNG): schema.O30_RNG_SHORT,
        (V30, REFERENCE, XSD): schema.O30_XSD_REFERENCE,
        (V30, REFERENCE, RNG): schema.O30_RNG_REFERENCE,
        (V31, SHORT, XSD): schema.O31_XSD_SHORT,
        (V31, SHORT, RNG): schema.O31_RNG_SHORT,
        (V31, REFERENCE, XSD): schema.O31_XSD_REFERENCE,
        (V31, REFERENCE, RNG): schema.O31_RNG_REFERENCE,
    }

    SCHEMA_TYPE_PARSER_MAP = {
        XSD: etree.XMLSchema,
        RNG: etree.RelaxNG,
    }

    @classmethod
    def from_tree(cls, tree):
        """
        Construct OnixMeta from an ElementTree.

        :param tree etree.ElementTree: LXML Parsed ONIX data
        :return OnixMeta: Initialized OnixMeta instance
        """
        root = tree.getroot()

        if root.tag.endswith("ONIXmessage"):
            onix_style = cls.SHORT
        elif root.tag.endswith("ONIXMessage"):
            onix_style = cls.REFERENCE
        else:
            raise OnixError("Bad root element: %s" % root.tag)

        onix_version = root.attrib.get("release")
        if onix_version is None:
            log.warning("No release attribute on root element. Try namespace.")
            try:
                if cls.V21 in list(root.nsmap.values())[0]:
                    onix_version = cls.V21
                elif cls.V30 in list(root.nsmap.values())[0]:
                    onix_version = cls.V30
                else:
                    raise OnixError("Could not determin ONIX version.")
            except IndexError:
                raise OnixError("No release attribute and no Namespace :(")

        namespaces = list(root.nsmap.values())
        return cls(
            xml_version=tree.docinfo.xml_version,
            xml_encoding=tree.docinfo.encoding,
            onix_version=onix_version,
            onix_style=onix_style,
            namespaces=namespaces,
        )

    @classmethod
    def from_file(cls, infile):
        """
        Contruct OnixMeta from an infile.

        :param infile: File or Path to file
        :type infile: file or str
        :return OnixMeta: Initialized OnixMeta instance
        """
        tree = etree.parse(infile)
        return cls.from_tree(tree)

    def get_ns_string(self):
        if self.onix_version == self.V21:
            tpl = "http://www.editeur.org/onix/2.1/%s"
        elif self.onix_version == self.V30:
            tpl = "http://ns.editeur.org/onix/3.0/%s"
        elif self.onix_version == self.V31:
            tpl = "http://ns.editeur.org/onix/3.1/%s"
        else:
            raise OnixError("Unsupported ONIX version: %s" % self.onix_version)
        return tpl % self.onix_style

    def get_schema_file(self, schema_type=XSD):
        key = self.onix_version, self.onix_style, schema_type
        try:
            return self.SCHEMA_MAP[key]
        except KeyError:
            raise OnixError("Found no {2} schema for ONIX {0} {1}".format(*key))


_BaseMessage = namedtuple("Message", "level validator location message error_type")


class Message(_BaseMessage):
    """
    A Validation message representing a single error condition.

    :param str level: Error level
    :param str validator: The validator that raised the error
    :param str location: Location of error (filename:line:column)
    :param str message: Description of the error condiction
    :param str error_type: Type of error
    """

    def __str__(self):
        return " | ".join(self._asdict().values())

    @property
    def short(self):
        """Short string representation of message"""
        return "{m.level} - {m.validator} - {m.location} - {m.message}".format(m=self)

    @classmethod
    def from_logentry(cls, logentry, filename=""):
        """Instanciate Message from lxml LogEntry object

        :param _LogEntry logentry: Validatation error from LXML
        :param str filename: Optional filename to prefix error location
        :return Message:
        """
        location = "%s:%s:%s" % (filename, logentry.line, logentry.column)
        message = logentry.message or ""
        message = re.sub("({.*?})", "", message)

        return cls(
            level=logentry.level_name,
            validator=logentry.domain_name,
            location=location,
            message=message,
            error_type=logentry.type_name,
        )

    @classmethod
    def from_exception(cls, exc, filename=""):
        """

        :param Exception exc:
        :param str filename: Optional filename to prefix error location
        :return Message:
        """
        return cls(
            level="CRITICAL",
            validator="ONIXCHECK",
            location=filename,
            message=exc.message,
            error_type="EXCEPTION",
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function
import sys
import logging
from argparse import ArgumentParser, FileType
from onixcheck import validate, __version__
from onixcheck.exeptions import get_logger
from onixcheck.utils import iter_files
from six.moves import getcwd

DEFAULT_EXTENSIONS = ("xml", "onx", "onix")
log = get_logger()


def create_parser():
    """Create a custom commandline parser.
    :rtype ArgumentParser:
    """
    parser = ArgumentParser(
        prog="onixcheck", description="Onixcheck v%s - Validate your metadata" % __version__
    )

    # Arguments
    parser.add_argument("infile", nargs="?", type=FileType(mode="rb"), help="Path to Onix file to validate")

    # Options
    parser.add_argument("-p", "--path", help="Path to folder with ONIX files for batch validation")
    parser.add_argument("-s", "--schemas", nargs="*", help="Custom validation schemas")
    parser.add_argument("-e", "--ext", nargs="*", help="File Extensions to validate. Default: xml onx onix")
    parser.add_argument("-r", "--recursive", action="store_true", help="Recurse into subfolders")
    parser.add_argument("-d", "--debug", action="store_true", help="Show debug information")
    parser.add_argument("-v", "--version", action="store_true", help="Show version information")

    # Defaults
    parser.set_defaults(ext=DEFAULT_EXTENSIONS)
    return parser


def main(argv=None):
    """Command line app main function.

    :param list | None argv: Overrides command options (for libuse or testing)
    """
    parser = create_parser()

    args = parser.parse_args() if argv is None else parser.parse_args(argv)

    if args.version:
        print("onixcheck version %s" % __version__)
        return 0

    schemas = ("xsd",) if not args.schemas else tuple(args.schemas)

    if args.debug:
        logging.basicConfig(level=logging.DEBUG, format="%(asctime)s - %(levelname)s - %(name)s - %(message)s")
        print("DEBUG logging enabled.")

    try:
        import win_unicode_console

        win_unicode_console.enable()
        log.debug("Running with win-unicode-console patch")
    except Exception:
        pass

    log.debug("TYPE of path: %s" % type(args.path))
    # validate current working dir
    if not args.infile and not args.path:
        args.path = getcwd()
        log.debug("NEW TYPE of path: %s" % type(args.path))

    all_valid = True

    if args.infile:
        log.debug("TYPE of infile.name: %s" % type(args.infile.name))
        print("Validating: %s" % args.infile.name)
        messages = validate(args.infile, schemas)
        is_valid = messages == []
        if is_valid:
            print("VALID - No errors found")
        else:
            print("INVALID - errors found:", file=sys.stderr)
            all_valid = False
            for msg in messages:
                if args.debug:
                    print(msg.__str__(), file=sys.stderr)
                else:
                    print(msg.short, file=sys.stderr)

    if args.path:
        tree_or_dir = "tree" if args.recursive else "dir"
        print()
        print("Validating all files in %s %s" % (tree_or_dir, args.path))

        for onix_file_path in iter_files(args.path, args.ext, args.recursive):
            print()
            print("Validating: %s" % onix_file_path)
            with open(onix_file_path, "rb") as onix_file:
                messages = validate(onix_file, schemas)
                is_valid = messages == []

            if is_valid:
                print("VALID - No errors found")
            else:
                print("INVALID - errors found:", file=sys.stderr)
                all_valid = False
                for msg in messages:
                    if args.debug:
                        print(msg.__str__(), file=sys.stderr)
                    else:
                        print(msg.short, file=sys.stderr)
    if all_valid:
        return 0
    else:
        return 1


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
import logging


class OnixError(Exception):
    pass


class NullHandler(logging.Handler):
    """Not in python 2.6 so we use our own"""

    def emit(self, record):
        pass


def get_logger(logger_name="onixcheck", add_null_handler=True):
    logger = logging.getLogger(logger_name)
    if add_null_handler:
        logger.addHandler(NullHandler())
    return logger



# -*- coding: utf-8 -*-
"""Generic or common utility functions"""

from __future__ import print_function, unicode_literals
from os.path import splitext, join

has_scandir = True

try:
    from scandir import scandir, walk
except ImportError:
    from os import walk
    from os import listdir as scandir
    from os.path import isdir

    has_scandir = False


def iter_files(root, exts=None, recursive=False):
    """
    Iterate over file paths within root filtered by specified extensions.

    :param str root: Root folder to start collecting files
    :param iterable exts: Restrict results to given file extensions
    :param bool recursive: Wether to walk the complete directory tree
    :rtype collections.Iterable[str]: absolute file paths with given extensions
    """

    if exts is not None:
        exts = set((x.lower() for x in exts))

    def matches(e):
        return (exts is None) or (e in exts)

    if recursive is False:
        for entry in scandir(root):
            if has_scandir:
                ext = splitext(entry.name)[-1].lstrip(".").lower()
                if entry.is_file() and matches(ext):
                    yield entry.path
            else:
                ext = splitext(entry)[-1].lstrip(".").lower()
                if not isdir(entry) and matches(ext):
                    yield join(root, entry)
    else:
        for root, folders, files in walk(root):
            for f in files:
                ext = splitext(f)[-1].lstrip(".").lower()
                if matches(ext):
                    yield join(root, f)



# -*- coding: utf-8 -*-
"""Paths to bundled Schema files"""

import sys
import os

if getattr(sys, "frozen", False):
    # we are running in a |PyInstaller| bundle
    basedir = sys._MEIPASS
else:
    # we are running in a normal Python environment
    basedir = os.path.dirname(__file__)

O21_XSD_SHORT = os.path.join(basedir, "xsd2.1", "ONIX_BookProduct_Release2.1_short.xsd")
O21_XSD_REFERENCE = os.path.join(basedir, "xsd2.1", "ONIX_BookProduct_Release2.1_reference.xsd")

O30_XSD_SHORT = os.path.join(basedir, "xsd3.0", "ONIX_BookProduct_3.0_short.xsd")
O30_XSD_REFERENCE = os.path.join(basedir, "xsd3.0", "ONIX_BookProduct_3.0_reference.xsd")

O30_RNG_SHORT = os.path.join(basedir, "rng3.0", "ONIX_BookProduct_3.0_short.rng")
O30_RNG_REFERENCE = os.path.join(basedir, "rng3.0", "ONIX_BookProduct_3.0_reference.rng")

O31_XSD_SHORT = os.path.join(basedir, "xsd3.1", "ONIX_BookProduct_3.1_short.xsd")
O31_XSD_REFERENCE = os.path.join(basedir, "xsd3.1", "ONIX_BookProduct_3.1_reference.xsd")

O31_RNG_SHORT = os.path.join(basedir, "rng3.1", "ONIX_BookProduct_3.1_short.rng")
O31_RNG_REFERENCE = os.path.join(basedir, "rng3.1", "ONIX_BookProduct_3.1_reference.rng")

GOOGLE_O30_YML_REFERENCE = os.path.join(basedir, "google", "google_onix_30.yaml")

BIBLON_O30_YML_REFERENCE = os.path.join(basedir, "biblon", "biblon_onix_30.yaml")



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from os.path import basename
from lxml.etree import XMLSyntaxError
from onixcheck.exeptions import OnixError
from onixcheck.models import OnixFile, Message
from onixcheck.onixfix import OnixFix
from onixcheck import schema


def validate(infile, schemas=("xsd",)):
    """Validate an ONIX file.

    :param infile: File or path to file
    :type infile: file or str
    :param schemas: Iterable with paths to custom valdation profiles
    :type  schemas: collections.Iterable[str]
    :return: List of `Message` objects (invalid ONIX) or empty list (valid ONIX)
    :rtype: list[Message]
    """
    if hasattr(infile, "name"):
        filename = basename(infile.name)
    else:
        filename = basename(infile)

    try:
        onix_file = OnixFile(infile)
    except (OnixError, XMLSyntaxError) as e:
        return [Message.from_exception(e, filename)]

    messages = []

    for s in schemas:
        if s in (
            "xsd",
            "rng",
        ):
            try:
                validator = onix_file.get_validator(s)
            except OnixError as e:
                messages.append(Message.from_exception(e, filename))
                continue
            validator(onix_file.xml_tree())
            errors = validator.error_log
            msg = Message.from_logentry
            messages.extend([msg(err, filename) for err in errors])

    for s in schemas:
        if s in ("google", "biblon"):
            if s == "google":
                profile = schema.GOOGLE_O30_YML_REFERENCE
            elif s == "biblon":
                profile = schema.BIBLON_O30_YML_REFERENCE

            validator = OnixFix(infile, profile)
            validator.validate()
            messages.extend(validator.errors)

    return messages
